(*
    http://www.microsoft.com/typography/otspec/otff.htm
    The OpenType Font File
*)

unit otff;

interface

uses
    Windows, SysUtils, Classes;

procedure GetFontNameFromTTC(lpszFilePath: string; Results: TStrings);

implementation

type
    USHORT = Word;
    ULONG = Longword;
    Tag = array[0..3] of char;
    Fixed = ULONG;              //  ۂ 32bit fixed-point number(16.16)

type
    //  Version 1.0  TTC wb_
    TTC_HEADER_VERSION_10 = packed record
        TTCTag      : Tag;                  //  TrueType Collection ID string: 'ttcf'
        Version     : ULONG;                //  Version of the TTC Header (1.0), 0x00010000
        numFonts    : ULONG;                //  numFonts Number of fonts in TTC
        OffsetTable : array[0..0] of ULONG; //  ۂ array[0..numFonts-1] of ULONG;
                                            //  Array of offsets to the OffsetTable
                                            //  for each font from the beginning of the file
    end;

    //  Version 2.0  TTC wb_
    //  tHg~ŁAVOl`Ȃ񂩂ǂł悢B
    //  擪 4tB[h͋ʂ̂߁AVer.2.0 ͎gpȂB

    TTC_HEADER_VERSION_20 = packed record
        TTCTag      : Tag;                  //  TrueType Collection ID string: 'ttcf'
        Version     : ULONG;                //  Version of the TTC Header (2.0), 0x00020000
        numFonts    : ULONG;                //  numFonts Number of fonts in TTC
        OffsetTable : array[0..0] of ULONG; //  ۂ array[0..numFonts-1] of ULONG;
                                            //  Array of offsets to the OffsetTable
                                            //  for each font from the beginning of the file
        ulDsigTag   : ULONG;                //  Tag indicating that a DSIG table exists,
                                            //  0x44534947 ('DSIG') (null if no signature)
        ulDsigLength: ULONG;                //  The length (in bytes) of the DSIG table
                                            //  (null if no signature)
        ulDsigOffset: ULONG;                //  The offset (in bytes) of the DSIG table
                                            //  from the beginning of the TTC file
                                            //  (null if no signature)
    end;

    //  Offset Table
    TTC_OFFSET_TABLE = packed record
        sfntversion  : Fixed;           //  0x00010000 for version 1.0.
        numTables    : USHORT;          //  Number of tables.
        searchRange  : USHORT;          //  (Maximum power of 2 <= numTables) x 16.
        entrySelector: USHORT;          //  Log2(maximum power of 2 <= numTables).
        rangeShift   : USHORT;          //  NumTables x 16-searchRange.
    end;

    //  Table Directory
    TTC_TABLE_DIRECTORY = packed record
        tag     : array[0..3] of char;  //  4 -byte identifier.
        checkSum: ULONG;                //  CheckSum for this table.
        offset  : ULONG;                //  Offset from beginning of TrueType font file.
        length  : ULONG;                //  Length of this table.
    end;

    //  Naming Table
    TTC_NAMING_TABLE = packed record

    end;

    //  ----------------------------------------- from CodeGuruC3659.pas
    //  Header of the names table
    TT_NAME_TABLE_HEADER = packed record
        uFSelector    : USHORT;	    //  format selector. Always 0
        uNRCount      : USHORT;	    //  Name Records count
        uStorageOffset: USHORT;     //  Offset for strings storage,
                                    //  from start of the table
    end;

    //  ----------------------------------------- from CodeGuruC3659.pas
    //  Records in the names table
    TT_NAME_RECORD = packed record
        uPlatformID   : USHORT;
        uEncodingID   : USHORT;
        uLanguageID   : USHORT;
        uNameID       : USHORT;
        uStringLength : USHORT;
        uStringOffset : USHORT;     //  from start of storage area
    end;

//  ---------------------------------------------------------------------------
//  #define SWAPWORD(x) MAKEWORD(HIBYTE(x), LOBYTE(x))
function    SWAPWORD(x: Word): Word;
begin
    Result := MakeWord(HiByte(x), LoByte(x));
end;

//  ---------------------------------------------------------------------------
//  #define SWAPLONG(x) MAKELONG(SWAPWORD(HIWORD(x)),SWAPWORD(LOWORD(x)))
function    SWAPLONG(x: Cardinal): Cardinal;
begin
    Result := MakeLong(SWAPWORD(HiWord(x)), SWAPWORD(LoWord(x)));
end;

//  ---------------------------------------------------------------------------
procedure GetFontNameFromTTC(lpszFilePath: string; Results: TStrings);
var
    F: TFileStream;
    ttc_header_ver10: TTC_HEADER_VERSION_10;
    FontCount: integer;
    Offsets: array[0..256] of ULONG;
    i, j: integer;
    ttcOffsetTable: TTC_OFFSET_TABLE;
    TableCount: integer;
    ttcTableDir: TTC_TABLE_DIRECTORY;
    csTemp: string;
    ttNTHeader: TT_NAME_TABLE_HEADER;
    ttRecord: TT_NAME_RECORD;
    bFound: Boolean;
    nPos, nPos2: integer;
    Buf: array[0..1024] of char; //  ...
begin
    //  lpszFilePath is the path to our font file
    F := TFileStream.Create(lpszFilePath, fmOpenRead or fmShareDenyWrite);
    try
        //  wb_Ǎ
        F.Read(ttc_header_ver10, SizeOf(ttc_header_ver10));

        //  TrueType Collection mF
        if (ttc_header_ver10.TTCTag <> 'ttcf') then exit;

        //  o[W Big Endian -> Little Endian
        ttc_header_ver10.Version := SWAPLONG(ttc_header_ver10.Version);

        //  o[W(sv)
        if (ttc_header_ver10.Version = $00010000) then
        begin
            //  Version 1.0
        end
        else if (ttc_header_ver10.Version = $00020000) then
        begin
            //  Version 2.0
        end;

        //  ܂܂tHg̐𒲂ׂ
        ttc_header_ver10.numFonts := SWAPLONG(ttc_header_ver10.numFonts);
        FontCount := ttc_header_ver10.numFonts;

        //  etHg̃ItZbgʒui[zNA
        FillChar(Offsets, SizeOf(Offsets), 0);

        //  Xg[߂
        F.Seek(- SizeOf(ULONG), soFromCurrent);

        //  ItZbgʒuǂݍ
        F.Read(Offsets, SizeOf(ULONG) * FontCount);

        //  ItZbgʒuGfBAϊ
        for i := 0 to FontCount - 1 do
        begin
            Offsets[i] := SWAPLONG(Offsets[i]);
        end;

        for i := 0 to FontCount - 1 do
        begin
            //  Xg[𑁑
            F.Seek(Offsets[i], soFromBeginning);

            //  ItZbge[uǂݍ
            F.Read(ttcOffsetTable, SizeOf(ttcOffsetTable));

            //  GfBAϊ
            ttcOffsetTable.numTables := SWAPWORD(ttcOffsetTable.numTables);
            TableCount := ttcOffsetTable.numTables;

            //  [name]^O邩ǂtO
            bFound := false;

            //  炷 Table Directory ͂܂
            for j := 0 to TableCount - 1 do
            begin
                //  Ǎ
                F.Read(ttcTableDir, SizeOf(ttcTableDir));

                //  [name]^O
                if (ttcTableDir.tag = 'name') then
                begin
                    bFound := true;
                    break;
                end;
            end;

            //  Ȃ玟
            if not bFound then continue;

            //  GfBAϊ
            ttcTableDir.checkSum := SWAPLONG(ttcTableDir.checkSum);
            ttcTableDir.offset := SWAPLONG(ttcTableDir.offset);
            ttcTableDir.length := SWAPLONG(ttcTableDir.length);

            //  ݈ʒuL^
            nPos := F.Position;

            //  wʒu܂ňړ
            F.Seek(ttcTableDir.offset, soFromBeginning);

            //  ȍ~͂ււB
            //  ̍\̓Ǎ
            F.Read(ttNTHeader, SizeOf(ttNTHeader));

            //  GfBAϊ
            ttNTHeader.uNRCount       := SWAPWORD(ttNTHeader.uNRCount);
            ttNTHeader.uStorageOffset := SWAPWORD(ttNTHeader.uStorageOffset);

            //  ̃R[h̐JԂ
            for j := 0 to ttNTHeader.uNRCount - 1 do
            begin
                F.Read(ttRecord, SizeOf(TT_NAME_RECORD));
                ttRecord.uNameID := SWAPWORD(ttRecord.uNameID);

                //  4 ̓tl[
                if (ttRecord.uNameID = 6) then
                begin
                    ttRecord.uStringLength := SWAPWORD(ttRecord.uStringLength);
                    ttRecord.uStringOffset := SWAPWORD(ttRecord.uStringOffset);

                    //  ݈ʒuL^
                    nPos2 := F.Position;
                    F.Seek(ttcTableDir.offset
                           + ttRecord.uStringOffset
                           + ttNTHeader.uStorageOffset,
                           soFromBeginning);

                    FillChar(Buf, SizeOf(Buf), 0);
                    F.Read(Buf, ttRecord.uStringLength);
                    csTemp := string(Buf);

                    //  Ύ
                    if (csTemp <> '') then
                    begin
                        Results.Add(String(Buf));
                        break;
                    end;
                    F.Seek(nPos2, soFromBeginning);
                end;
            end;

            //  ߂
            F.Seek(nPos, soFromBeginning);
        end;

    finally
        F.Free;
    end;
end;

end.

